(* 
 *  OCaml datatypes that we use to represent
 *  a simple, imperative programming language. 
 *)

(*
 *
 * Recall the BNF grammar for this simple language:
 *
 *  <exp> ::= 
 *         |  <X>
 *         |  <exp> + <exp>
 *         |  <exp> * <exp>
 *         |  <exp> < <exp>
 *         |  <integer contant>
 *         |  (<exp>)
 *
 *  (3 + 4) * 5
 *
 *  <cmd> ::= 
 *         |  skip
 *         |  <X> = <exp>
 *         |  ifNZ <exp> { <cmd> } else { <cmd> }
 *         |  whileNZ <exp> { <cmd> }
 *         |  <cmd>; <cmd>
 *
 *)

type var = string

type exp =
  | Var of var
  | Add of exp * exp
  | Mul of exp * exp
  | Lt  of exp * exp
  | Lit of int

type cmd =
  | Skip
  | Assn of var * exp
  | IfNZ of exp * cmd * cmd
  | WhileNZ of exp * cmd
  | Seq of cmd * cmd

(* Abstract Syntax (AST) for the concrete syntax:

        X = 6;
	ANS = 1;
	whileNZ (x) {
  		ANS = ANS * X;
  		X = X + -1;
	} 
 *)
let factorial : cmd =
  let x = "X" in
  let ans = "ANS" in
  Seq (Assn(x, Lit 6),
       Seq(Assn(ans, Lit 1),
           WhileNZ(Var x,
                   Seq(Assn(ans, Mul(Var ans, Var x)),
                       Assn(x, Add(Var x, Lit(-1)))))))


type state = string -> int

let init_state = fun x -> 0

let update (s:state) (x:var) (v:int) : state =
  fun (y:var) ->
    if x = y then v else s y

let rec interpret_exp (s:state) (e:exp) : int =
  begin match e with
    | Var x -> s x
    | Add(e1, e2) -> (interpret_exp s e1) + (interpret_exp s e2)
    | Mul(e1, e2) -> (interpret_exp s e1) * (interpret_exp s e2)
    | Lt(e1, e2) -> if (interpret_exp s e1) < (interpret_exp s e2) then 1 else 0
    | Lit i -> i
  end

let rec interpret_cmd (s:state) (c:cmd) : state =
  begin match c with
    | Skip -> s
    | Assn(x, e) -> update s x (interpret_exp s e)
    | IfNZ(e, c1, c2) ->
      if (interpret_exp s e) <> 0 then interpret_cmd s c1 else interpret_cmd s c2
    | WhileNZ(e, c) ->
      interpret_cmd s (IfNZ(e, Seq(c, WhileNZ(e, c)), Skip))
    | Seq(c1, c2) ->
      let s1 = interpret_cmd s c1 in
      interpret_cmd s1 c2
  end


let main () =
  let s_ans = interpret_cmd init_state factorial in
  Printf.printf "ANS = %d\n" (s_ans "ANS")

;; main ()

